import { messageTraversal, signalTraversal } from '../dsp-graph/traversers.js';
import macros from '../engine-javascript/compile/macros.js';
import macros$1 from '../engine-assemblyscript/compile/macros.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
/** Helper to get code macros from compile target. */
const getMacros = (target) => ({ javascript: macros, assemblyscript: macros$1 }[target]);
/** Helper to get node implementation or throw an error if not implemented. */
const getNodeImplementation = (nodeImplementations, nodeType) => {
    const nodeImplementation = nodeImplementations[nodeType];
    if (!nodeImplementation) {
        throw new Error(`node [${nodeType}] is not implemented`);
    }
    return {
        dependencies: [],
        ...nodeImplementation,
    };
};
/**
 * Build graph traversal for all nodes.
 * This should be exhaustive so that all nodes that are connected
 * to an input or output of the graph are declared correctly.
 * Order of nodes doesn't matter.
 */
const buildFullGraphTraversal = (graph) => {
    const nodesPullingSignal = Object.values(graph).filter((node) => !!node.isPullingSignal);
    const nodesPushingMessages = Object.values(graph).filter((node) => !!node.isPushingMessages);
    return Array.from(new Set([
        ...messageTraversal(graph, nodesPushingMessages),
        ...signalTraversal(graph, nodesPullingSignal),
    ]));
};
const getNodeImplementationsUsedInGraph = (graph, nodeImplementations) => Object.values(graph).reduce((nodeImplementationsUsedInGraph, node) => {
    if (node.type in nodeImplementationsUsedInGraph) {
        return nodeImplementationsUsedInGraph;
    }
    else {
        return {
            ...nodeImplementationsUsedInGraph,
            [node.type]: getNodeImplementation(nodeImplementations, node.type),
        };
    }
}, {});
/**
 * Build graph traversal for all signal nodes.
 */
const buildGraphTraversalSignal = (graph) => signalTraversal(graph, getGraphSignalSinks(graph));
const getGraphSignalSinks = (graph) => Object.values(graph).filter((node) => !!node.isPullingSignal);
const isGlobalDefinitionWithSettings = (globalCodeDefinition) => !(typeof globalCodeDefinition === 'function');

export { buildFullGraphTraversal, buildGraphTraversalSignal, getGraphSignalSinks, getMacros, getNodeImplementation, getNodeImplementationsUsedInGraph, isGlobalDefinitionWithSettings };
