import { isGlobalDefinitionWithSettings } from '../compile-helpers.js';
import { toNodes } from '../../dsp-graph/traversers.js';
import { Sequence } from '../../ast/declare.js';
import { commonsArrays, commonsWaitFrame } from '../../stdlib/commons.js';
import { core } from '../../stdlib/core.js';
import '../../stdlib/fs.js';
import { msg } from '../../stdlib/msg.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
var precompileDependencies = (precompilation) => {
    const { settings, variableNamesAssigner, precompiledCodeAssigner, } = precompilation;
    const dependencies = flattenDependencies([
        ...engineMinimalDependencies(),
        ..._collectDependenciesFromTraversal(precompilation),
    ]);
    // Flatten and de-duplicate all the module's dependencies
    precompiledCodeAssigner.dependencies.ast = instantiateAndDedupeDependencies(settings, dependencies, variableNamesAssigner.globs);
    // Collect and attach imports / exports info
    precompiledCodeAssigner.dependencies.exports = collectAndDedupeExports(settings.target, dependencies);
    precompiledCodeAssigner.dependencies.imports =
        collectAndDedupeImports(dependencies);
};
const instantiateAndDedupeDependencies = (settings, dependencies, globs) => {
    const context = {
        settings,
        globs,
    };
    return Sequence(dependencies
        .map((codeDefinition) => isGlobalDefinitionWithSettings(codeDefinition)
        ? codeDefinition.codeGenerator(context)
        : codeDefinition(context))
        .reduce((astElements, astElement) => astElements.every((otherElement) => !_deepEqual(otherElement, astElement))
        ? [...astElements, astElement]
        : astElements, []));
};
const engineMinimalDependencies = () => [
    core,
    commonsArrays,
    commonsWaitFrame,
    msg,
];
const collectAndDedupeExports = (target, dependencies) => dependencies
    .filter(isGlobalDefinitionWithSettings)
    .reduce((exports, codeDefinition) => codeDefinition.exports
    ? [
        ...exports,
        ...codeDefinition.exports.filter((xprt) => (!xprt.targets ||
            xprt.targets.includes(target)) &&
            exports.every((otherExport) => xprt.name !== otherExport.name)),
    ]
    : exports, []);
const collectAndDedupeImports = (dependencies) => dependencies
    .filter(isGlobalDefinitionWithSettings)
    .reduce((imports, codeDefinition) => codeDefinition.imports
    ? [
        ...imports,
        ...codeDefinition.imports.filter((imprt) => imports.every((otherImport) => imprt.name !== otherImport.name)),
    ]
    : imports, []);
const flattenDependencies = (dependencies) => dependencies.flatMap((codeDefinition) => {
    if (isGlobalDefinitionWithSettings(codeDefinition) &&
        codeDefinition.dependencies) {
        return [
            ...flattenDependencies(codeDefinition.dependencies),
            codeDefinition,
        ];
    }
    else {
        return [codeDefinition];
    }
});
const _collectDependenciesFromTraversal = ({ graph, precompiledCodeAssigner, }) => {
    return toNodes(graph, precompiledCodeAssigner.graph.fullTraversal)
        .reduce((definitions, node) => {
        const precompiledNode = precompiledCodeAssigner.nodes[node.id];
        const precompiledNodeImplementation = precompiledCodeAssigner.nodeImplementations[precompiledNode.nodeType];
        return [
            ...definitions,
            ...(precompiledNodeImplementation.nodeImplementation
                .dependencies || []),
        ];
    }, []);
};
const _deepEqual = (ast1, ast2) => 
// This works but this flawed cause {a: 1, b: 2} and {b: 2, a: 1}
// would compare to false.
JSON.stringify(ast1) === JSON.stringify(ast2);

export { collectAndDedupeExports, collectAndDedupeImports, precompileDependencies as default, engineMinimalDependencies, flattenDependencies, instantiateAndDedupeDependencies };
