import { signalTraversal, toNodes, listSinkNodes, listSinkConnections } from '../../dsp-graph/traversers.js';
import { endpointsEqual } from '../../dsp-graph/graph-helpers.js';
import { buildGraphTraversalSignal } from '../compile-helpers.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
const precompileColdDspGroup = ({ graph, variableNamesAssigner, precompiledCodeAssigner, }, dspGroup, groupId) => {
    precompiledCodeAssigner.graph.coldDspGroups[groupId] = {
        functionName: variableNamesAssigner.coldDspGroups[groupId],
        dspGroup,
        sinkConnections: buildGroupSinkConnections(graph, dspGroup),
    };
};
const buildHotDspGroup = ({ graph }, parentDspGroup, coldDspGroups) => ({
    traversal: coldDspGroups.reduce((traversal, coldDspGroup) => removeNodesFromTraversal(traversal, coldDspGroup.traversal), buildGraphTraversalSignal(graph)),
    outNodesIds: parentDspGroup.outNodesIds,
});
const buildColdDspGroups = (precompilation, parentDspGroup) => 
// Go through all nodes in the signal traversal, and find groups of signal nodes
// that can be cached / computed only when needed (cold dsp), outside
// of the main dsp loop (hot dsp). We proceed in the following way :
//
// 1. Find single flow dsp groups, i.e. a cold node and all its sources,
//      and their sources, etc. as long as all nodes are cold.
// 2. some of these single flow dsp groups might be connected to each other,
//      therefore we need to merge them.
// 3. for each merged group consolidate the signal traversal, therefore
//      fixing the order in which nodes are visited and removing potential duplicates.
//
// e.g. :
//
//      [  c1  ]
//         |\________
//         |         |
//      [  c2  ]  [  c4  ]
//         |         |
//      [  c3  ]  [  c5  ]  <- out nodes of cold the cold dsp group
//         |         |
//      [  h1  ]  [  h2  ]  <- hot nodes
//
// In the graph above, [c1, c2, c3, c4, c5] constitute a cold dsp group.
// 1. We start by finding 2 single flow dsp groups : [c3, c2, c1] and [c5, c4, c1]
// 2. We detect that these 2 groups are connected, so we merge them into one group : [c3, c2, c1, c5, c4, c1]
// 3. ...
_buildSingleFlowColdDspGroups(precompilation, parentDspGroup)
    // Combine all connected single flow dsp groups.
    .reduce((dspGroups, singleFlowDspGroup) => {
    const groupToMergeInto = dspGroups.find((otherGroup) => otherGroup.traversal.some((nodeId) => singleFlowDspGroup.traversal.includes(nodeId)));
    if (groupToMergeInto) {
        return [
            ...dspGroups.filter((dspGroup) => dspGroup !== groupToMergeInto),
            // Merging here is incomplete, we don't recompute the traversal
            // and don't remove the duplicate nodes until all groups are combined.
            {
                traversal: [
                    ...groupToMergeInto.traversal,
                    ...singleFlowDspGroup.traversal,
                ],
                outNodesIds: [
                    ...groupToMergeInto.outNodesIds,
                    ...singleFlowDspGroup.outNodesIds,
                ],
            },
        ];
    }
    else {
        return [...dspGroups, singleFlowDspGroup];
    }
}, [])
    // Compute the signal traversal, therefore fixing the order in which
    // nodes are visited and removing potential duplicates.
    .map((dspGroup) => ({
    traversal: signalTraversal(precompilation.graph, toNodes(precompilation.graph, dspGroup.outNodesIds)),
    outNodesIds: dspGroup.outNodesIds,
}));
const _buildSingleFlowColdDspGroups = (precompilation, parentDspGroup) => toNodes(precompilation.graph, parentDspGroup.traversal)
    .reduce((dspGroups, node) => {
    // If one of `node`'s sinks is a also cold, then `node` is not the
    // out node of a dsp group.
    if (!_isNodeDspCold(precompilation, node) ||
        listSinkNodes(precompilation.graph, node, 'signal')
            .every((sinkNode) => _isNodeDspCold(precompilation, sinkNode))) {
        return dspGroups;
    }
    // We need to check that all upstream nodes are also cold.
    let areAllSourcesCold = true;
    const dspGroup = {
        outNodesIds: [node.id],
        traversal: signalTraversal(precompilation.graph, [node], (sourceNode) => {
            areAllSourcesCold =
                areAllSourcesCold &&
                    _isNodeDspCold(precompilation, sourceNode);
            return areAllSourcesCold;
        }),
    };
    if (areAllSourcesCold) {
        return [...dspGroups, dspGroup];
    }
    else {
        return dspGroups;
    }
}, []);
const buildInlinableDspGroups = (precompilation, parentDspGroup) => toNodes(precompilation.graph, parentDspGroup.traversal)
    .reduce((dspGroups, node) => {
    const sinkNodes = listSinkNodes(precompilation.graph, node, 'signal');
    // We're looking for the out node of an inlinable dsp group.
    if (_isNodeDspInlinable(precompilation, node) &&
        // If node is the out node of its parent dsp group, then its not inlinable,
        // because it needs to declare output variables.
        !parentDspGroup.outNodesIds.includes(node.id) &&
        // If `node`'s sink is itself inlinable, then `node` is not the out node.
        (!_isNodeDspInlinable(precompilation, sinkNodes[0]) ||
            // However, if `node`'s sink is also is the out node of the parent group,
            // then it can't actually be inlined, so `node` is the out node.
            parentDspGroup.outNodesIds.includes(sinkNodes[0].id))) {
        return [
            ...dspGroups,
            {
                traversal: signalTraversal(precompilation.graph, [node], (sourceNode) => _isNodeDspInlinable(precompilation, sourceNode)),
                outNodesIds: [node.id],
            },
        ];
    }
    else {
        return dspGroups;
    }
}, []);
const isNodeInsideGroup = (dspGroup, nodeId) => dspGroup.traversal.includes(nodeId);
const findColdDspGroupFromSink = (coldDspGroupMap, sink) => Object.values(coldDspGroupMap).find(({ sinkConnections }) => sinkConnections.find(([_, otherSink]) => endpointsEqual(otherSink, sink)));
const buildGroupSinkConnections = (graph, dspGroup) => toNodes(graph, dspGroup.outNodesIds)
    // Get a flat list of all the sink connections of the out nodes.
    .flatMap((outNode) => listSinkConnections(outNode, 'signal'));
const removeNodesFromTraversal = (traversal, toRemove) => traversal.filter((nodeId) => !toRemove.includes(nodeId));
const _isNodeDspCold = ({ precompiledCodeAssigner }, node) => {
    const precompiledNode = precompiledCodeAssigner.nodes[node.id];
    const precompiledNodeImplementation = precompiledCodeAssigner.nodeImplementations[precompiledNode.nodeType];
    return precompiledNodeImplementation.nodeImplementation.flags
        ? !!precompiledNodeImplementation.nodeImplementation.flags
            .isPureFunction
        : false;
};
const _isNodeDspInlinable = ({ precompiledCodeAssigner }, node) => {
    const sinks = listSinkConnections(node, 'signal')
        .map(([_, sink]) => sink)
        // De-duplicate sinks
        .reduce((dedupedSinks, sink) => {
        if (dedupedSinks.every((otherSink) => !endpointsEqual(otherSink, sink))) {
            return [...dedupedSinks, sink];
        }
        else {
            return dedupedSinks;
        }
    }, []);
    const precompiledNode = precompiledCodeAssigner.nodes[node.id];
    const precompiledNodeImplementation = precompiledCodeAssigner.nodeImplementations[precompiledNode.nodeType];
    return (!!precompiledNodeImplementation.nodeImplementation.flags &&
        !!precompiledNodeImplementation.nodeImplementation.flags.isDspInline &&
        sinks.length === 1);
};

export { _buildSingleFlowColdDspGroups, _isNodeDspInlinable, buildColdDspGroups, buildGroupSinkConnections, buildHotDspGroup, buildInlinableDspGroups, findColdDspGroupFromSink, isNodeInsideGroup, precompileColdDspGroup, removeNodesFromTraversal };
