import { getNode } from '../../dsp-graph/getters.js';
import { getNodeImplementation } from '../compile-helpers.js';
import { Assigner, ProtectedIndex } from '../proxies.js';
import { ast, Sequence } from '../../ast/declare.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
// ---------------------------- VariableNamesIndex ---------------------------- //
const VariableNamesAssigner = ({ input: context, variableNamesIndex, }) => Assigner(_VariableNamesAssignerSpec, variableNamesIndex, context);
const createVariableNamesIndex = (precompilationInput) => Assigner.ensureValue({}, _VariableNamesAssignerSpec, precompilationInput);
const createGlobsVariableNames = () => ({
    iterFrame: 'F',
    frame: 'FRAME',
    blockSize: 'BLOCK_SIZE',
    sampleRate: 'SAMPLE_RATE',
    output: 'OUTPUT',
    input: 'INPUT',
    nullMessageReceiver: 'SND_TO_NULL',
    nullSignal: 'NULL_SIGNAL',
    emptyMessage: 'EMPTY_MESSAGE',
});
const _VariableNamesAssignerSpec = Assigner.Interface({
    nodes: Assigner.Index((nodeId) => Assigner.Interface({
        signalOuts: Assigner.Index((portletId) => Assigner.Literal(() => _name('N', nodeId, 'outs', portletId))),
        messageSenders: Assigner.Index((portletId) => Assigner.Literal(() => _name('N', nodeId, 'snds', portletId))),
        messageReceivers: Assigner.Index((portletId) => Assigner.Literal(() => _name('N', nodeId, 'rcvs', portletId))),
        state: Assigner.LiteralDefaultNull(() => _name('N', nodeId, 'state')),
    })),
    nodeImplementations: Assigner.Index((nodeType, { nodeImplementations }) => {
        const nodeImplementation = getNodeImplementation(nodeImplementations, nodeType);
        const nodeTypePrefix = (nodeImplementation.flags
            ? nodeImplementation.flags.alphaName
            : null) || nodeType;
        return Assigner.Index((name) => Assigner.Literal(() => _name('NT', nodeTypePrefix, name)));
    }),
    globs: Assigner.Literal(createGlobsVariableNames),
    globalCode: Assigner.Index((ns) => Assigner.Index((name) => Assigner.Literal(() => _name('G', ns, name)))),
    io: Assigner.Interface({
        messageReceivers: Assigner.Index((nodeId) => Assigner.Index((inletId) => Assigner.Literal(() => _name('IORCV', nodeId, inletId)))),
        messageSenders: Assigner.Index((nodeId) => Assigner.Index((outletId) => Assigner.Literal(() => _name('IOSND', nodeId, outletId)))),
    }),
    coldDspGroups: Assigner.Index((groupId) => Assigner.Literal(() => _name('COLD', groupId))),
});
// ---------------------------- PrecompiledCode ---------------------------- //
const PrecompiledCodeAssigner = ({ input: context, precompiledCode, }) => Assigner(_PrecompiledCodeAssignerSpec, precompiledCode, context);
const createPrecompiledCode = (precompilationInput) => Assigner.ensureValue({}, _PrecompiledCodeAssignerSpec, precompilationInput);
const _PrecompiledCodeAssignerSpec = Assigner.Interface({
    graph: Assigner.Literal((_, path) => ({
        fullTraversal: [],
        hotDspGroup: {
            traversal: [],
            outNodesIds: [],
        },
        coldDspGroups: ProtectedIndex({}, path),
    })),
    nodeImplementations: Assigner.Index((nodeType, { nodeImplementations }) => Assigner.Literal(() => ({
        nodeImplementation: getNodeImplementation(nodeImplementations, nodeType),
        stateClass: null,
        core: null,
    })), (_, path) => ProtectedIndex({}, path)),
    nodes: Assigner.Index((nodeId, { graph }) => Assigner.Literal(() => ({
        nodeType: getNode(graph, nodeId).type,
        messageReceivers: {},
        messageSenders: {},
        signalOuts: {},
        signalIns: {},
        initialization: ast ``,
        dsp: {
            loop: ast ``,
            inlets: {},
        },
        state: null,
    })), (_, path) => ProtectedIndex({}, path)),
    dependencies: Assigner.Literal(() => ({
        imports: [],
        exports: [],
        ast: Sequence([]),
    })),
    io: Assigner.Interface({
        messageReceivers: Assigner.Index((_) => Assigner.Literal((_, path) => ProtectedIndex({}, path)), (_, path) => ProtectedIndex({}, path)),
        messageSenders: Assigner.Index((_) => Assigner.Literal((_, path) => ProtectedIndex({}, path)), (_, path) => ProtectedIndex({}, path)),
    }),
});
// ---------------------------- MISC ---------------------------- //
const _name = (...parts) => parts.map(assertValidNamePart).join('_');
const assertValidNamePart = (namePart) => {
    const isInvalid = !VALID_NAME_PART_REGEXP.exec(namePart);
    if (isInvalid) {
        throw new Error(`Invalid variable name for code generation "${namePart}"`);
    }
    return namePart;
};
const VALID_NAME_PART_REGEXP = /^[a-zA-Z0-9_]+$/;

export { PrecompiledCodeAssigner, VariableNamesAssigner, assertValidNamePart, createGlobsVariableNames, createPrecompiledCode, createVariableNamesIndex };
