import { mapArray } from '../functional-helpers.js';
import { getNode, getOutlet, getInlet } from './getters.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * Simple helper to get a list of nodes from a traversal (which is simply node ids).
 */
const toNodes = (graph, traversal) => traversal.map((nodeId) => getNode(graph, nodeId));
const listSinkNodes = (graph, node, portletType) => _listSourceOrSinkNodes(node.sinks, getOutlet, graph, node, portletType);
const listSourceNodes = (graph, node, portletType) => _listSourceOrSinkNodes(node.sources, getInlet, graph, node, portletType);
const listSinkConnections = (node, portletType) => _listSourcesOrSinks(node.sinks, getOutlet, node, portletType);
const _listSourcesOrSinks = (sourcesOrSinks, portletGetter, node, portletType) => 
// We always put the `node` endpoint first, even if we're listing connections to sources,
// this allows mre genericity to the function
Object.entries(sourcesOrSinks).reduce((connections, [portletId, sourceOrSinkList]) => {
    const nodeEndpoint = { portletId, nodeId: node.id };
    const portlet = portletGetter(node, portletId);
    if (portlet.type === portletType || portletType === undefined) {
        return [
            ...connections,
            ...sourceOrSinkList.map((s) => [nodeEndpoint, s]),
        ];
    }
    else {
        return connections;
    }
}, []);
const _listSourceOrSinkNodes = (sourcesOrSinks, portletGetter, graph, node, portletType) => _listSourcesOrSinks(sourcesOrSinks, portletGetter, node, portletType)
    .reduce((sourceOrSinkNodeIds, [_, sourceOrSink]) => {
    if (!sourceOrSinkNodeIds.includes(sourceOrSink.nodeId)) {
        return [...sourceOrSinkNodeIds, sourceOrSink.nodeId];
    }
    else {
        return sourceOrSinkNodeIds;
    }
}, [])
    .map((nodeId) => getNode(graph, nodeId));
/**
 * Breadth first traversal for signal in the graph.
 * Traversal path is calculated by pulling incoming connections from
 * {@link nodesPullingSignal}.
 */
const signalTraversal = (graph, nodesPullingSignal, shouldContinue) => {
    const traversal = [];
    nodesPullingSignal.forEach((node) => _signalTraversalBreadthFirstRecursive(traversal, [], graph, node, shouldContinue));
    return traversal;
};
const _signalTraversalBreadthFirstRecursive = (traversal, currentPath, graph, node, shouldContinue) => {
    if (shouldContinue && !shouldContinue(node)) {
        return;
    }
    const nextPath = [...currentPath, node.id];
    listSourceNodes(graph, node, 'signal').forEach((sourceNode) => {
        if (currentPath.indexOf(sourceNode.id) !== -1) {
            return;
        }
        _signalTraversalBreadthFirstRecursive(traversal, nextPath, graph, sourceNode, shouldContinue);
    });
    if (traversal.indexOf(node.id) === -1) {
        traversal.push(node.id);
    }
};
/**
 * Breadth first traversal for signal in the graph.
 * Traversal path is calculated by pulling incoming connections from
 * {@link nodesPushingMessages}.
 */
const messageTraversal = (graph, nodesPushingMessages) => {
    const traversal = [];
    nodesPushingMessages.forEach((node) => {
        _messageTraversalDepthFirstRecursive(traversal, graph, node);
    });
    return traversal;
};
const _messageTraversalDepthFirstRecursive = (traversal, graph, node) => {
    if (traversal.indexOf(node.id) !== -1) {
        return;
    }
    traversal.push(node.id);
    listSinkNodes(graph, node, 'message').forEach((sinkNode) => {
        _messageTraversalDepthFirstRecursive(traversal, graph, sinkNode);
    });
};
/**
 * Remove dead sinks and sources in graph.
 * @param graphTraversal contains all nodes that are connected to
 * an input or output of the graph.
 */
const trimGraph = (graph, graphTraversal) => mapArray(Object.values(graph).filter((node) => graphTraversal.includes(node.id)), (node) => [
    node.id,
    {
        ...node,
        sources: removeDeadSources(node.sources, graphTraversal),
        sinks: removeDeadSinks(node.sinks, graphTraversal),
    },
]);
/**
 * When `node` has a sink node that is not connected to an end sink, that sink node won't be included
 * in the traversal, but will still appear in `node.sinks`.
 * Therefore, we need to make sure to filter `node.sinks` to exclude sink nodes that don't
 * appear in the traversal.
 */
const removeDeadSinks = (sinks, graphTraversal) => {
    const filteredSinks = {};
    Object.entries(sinks).forEach(([outletId, outletSinks]) => {
        const filteredOutletSinks = outletSinks.filter(({ nodeId: sinkNodeId }) => graphTraversal.includes(sinkNodeId));
        if (filteredOutletSinks.length) {
            filteredSinks[outletId] = filteredOutletSinks;
        }
    });
    return filteredSinks;
};
/**
 * Filters a node's sources to exclude source nodes that don't
 * appear in the traversal.
 */
const removeDeadSources = (sources, graphTraversal) => {
    const filteredSources = {};
    Object.entries(sources).forEach(([inletId, inletSources]) => {
        const filteredInletSources = inletSources.filter(({ nodeId: sourceNodeId }) => graphTraversal.includes(sourceNodeId));
        if (filteredInletSources.length) {
            filteredSources[inletId] = filteredInletSources;
        }
    });
    return filteredSources;
};

export { listSinkConnections, listSinkNodes, listSourceNodes, messageTraversal, removeDeadSinks, removeDeadSources, signalTraversal, toNodes, trimGraph };
