import { lowerListOfFloatArrays, liftString, readListOfFloatArrays } from './core-bindings.js';
import { liftMessage } from './msg-bindings.js';
import { updateWasmInOuts } from './engine-lifecycle-bindings.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
const createFsBindings = (rawModule, engineData) => ({
    sendReadSoundFileResponse: {
        type: 'proxy',
        value: (operationId, status, sound) => {
            let soundPointer = 0;
            if (sound) {
                soundPointer = lowerListOfFloatArrays(rawModule, engineData.bitDepth, sound);
            }
            rawModule.x_fs_onReadSoundFileResponse(operationId, status, soundPointer);
            updateWasmInOuts(rawModule, engineData);
        },
    },
    sendWriteSoundFileResponse: {
        type: 'proxy',
        value: rawModule.x_fs_onWriteSoundFileResponse,
    },
    sendSoundStreamData: {
        type: 'proxy',
        value: (operationId, sound) => {
            const soundPointer = lowerListOfFloatArrays(rawModule, engineData.bitDepth, sound);
            const writtenFrameCount = rawModule.x_fs_onSoundStreamData(operationId, soundPointer);
            updateWasmInOuts(rawModule, engineData);
            return writtenFrameCount;
        },
    },
    closeSoundStream: {
        type: 'proxy',
        value: rawModule.x_fs_onCloseSoundStream,
    },
    onReadSoundFile: { type: 'callback', value: () => undefined },
    onWriteSoundFile: { type: 'callback', value: () => undefined },
    onOpenSoundReadStream: { type: 'callback', value: () => undefined },
    onOpenSoundWriteStream: { type: 'callback', value: () => undefined },
    onSoundStreamData: { type: 'callback', value: () => undefined },
    onCloseSoundStream: { type: 'callback', value: () => undefined },
});
const createFsImports = (forwardReferences) => {
    let wasmImports = {
        i_fs_readSoundFile: (operationId, urlPointer, infoPointer) => {
            const url = liftString(forwardReferences.rawModule, urlPointer);
            const info = liftMessage(forwardReferences.rawModule, infoPointer);
            forwardReferences.modules.fs.onReadSoundFile(operationId, url, info);
        },
        i_fs_writeSoundFile: (operationId, soundPointer, urlPointer, infoPointer) => {
            const sound = readListOfFloatArrays(forwardReferences.rawModule, forwardReferences.engineData.bitDepth, soundPointer);
            const url = liftString(forwardReferences.rawModule, urlPointer);
            const info = liftMessage(forwardReferences.rawModule, infoPointer);
            forwardReferences.modules.fs.onWriteSoundFile(operationId, sound, url, info);
        },
        i_fs_openSoundReadStream: (operationId, urlPointer, infoPointer) => {
            const url = liftString(forwardReferences.rawModule, urlPointer);
            const info = liftMessage(forwardReferences.rawModule, infoPointer);
            // Called here because this call means that some sound buffers were allocated
            // inside the wasm module.
            updateWasmInOuts(forwardReferences.rawModule, forwardReferences.engineData);
            forwardReferences.modules.fs.onOpenSoundReadStream(operationId, url, info);
        },
        i_fs_openSoundWriteStream: (operationId, urlPointer, infoPointer) => {
            const url = liftString(forwardReferences.rawModule, urlPointer);
            const info = liftMessage(forwardReferences.rawModule, infoPointer);
            forwardReferences.modules.fs.onOpenSoundWriteStream(operationId, url, info);
        },
        i_fs_sendSoundStreamData: (operationId, blockPointer) => {
            const block = readListOfFloatArrays(forwardReferences.rawModule, forwardReferences.engineData.bitDepth, blockPointer);
            forwardReferences.modules.fs.onSoundStreamData(operationId, block);
        },
        i_fs_closeSoundStream: (...args) => forwardReferences.modules.fs.onCloseSoundStream(...args),
    };
    return wasmImports;
};

export { createFsBindings, createFsImports };
