import { createFsImports, createFsBindings } from './fs-bindings.js';
import { instantiateWasmModule } from './wasm-helpers.js';
import { createModule, getFloatArrayType } from '../../run/run-helpers.js';
import { createCommonsBindings } from './commons-bindings.js';
import { readMetadata, ioMsgSendersImports, createIoMessageReceiversBindings, createIoMessageSendersBindings, createEngineLifecycleBindings } from './engine-lifecycle-bindings.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
const createEngine = async (wasmBuffer) => {
    const { rawModule, engineData, forwardReferences } = await createRawModule(wasmBuffer);
    const engineBindings = await createBindings(rawModule, engineData, forwardReferences);
    return createModule(rawModule, engineBindings);
};
const createRawModule = async (wasmBuffer) => {
    // We need to read metadata before everything, because it is used by other initialization functions
    const metadata = await readMetadata(wasmBuffer);
    const forwardReferences = { modules: {} };
    const wasmImports = {
        ...createFsImports(forwardReferences),
        ...ioMsgSendersImports(forwardReferences, metadata),
    };
    const bitDepth = metadata.audioSettings.bitDepth;
    const arrayType = getFloatArrayType(bitDepth);
    const engineData = {
        metadata,
        wasmOutput: new arrayType(0),
        wasmInput: new arrayType(0),
        arrayType,
        bitDepth,
        blockSize: 0,
    };
    const wasmInstance = await instantiateWasmModule(wasmBuffer, {
        input: wasmImports,
    });
    const rawModule = wasmInstance.exports;
    return { rawModule, engineData, forwardReferences };
};
const createBindings = async (rawModule, engineData, forwardReferences) => {
    // Create bindings for core modules
    const commons = createModule(rawModule, createCommonsBindings(rawModule, engineData));
    const fs = createModule(rawModule, createFsBindings(rawModule, engineData));
    const io = {
        messageReceivers: createModule(rawModule, createIoMessageReceiversBindings(rawModule, engineData)),
        messageSenders: createModule(rawModule, createIoMessageSendersBindings(rawModule, engineData)),
    };
    // Update forward refs for use in Wasm imports
    forwardReferences.modules.fs = fs;
    forwardReferences.modules.io = io;
    forwardReferences.engineData = engineData;
    forwardReferences.rawModule = rawModule;
    // Build the full module
    return {
        ...createEngineLifecycleBindings(rawModule, engineData),
        metadata: { type: 'proxy', value: engineData.metadata },
        commons: { type: 'proxy', value: commons },
        fs: { type: 'proxy', value: fs },
        io: { type: 'proxy', value: io },
    };
};

export { createBindings, createEngine, createRawModule };
