import { countTo } from '../functional-helpers.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
const Var = (typeName, name, value) => _preventToString({
    astType: 'Var',
    name,
    type: typeName,
    value: value !== undefined ? _prepareVarValue(value) : undefined,
});
const ConstVar = (typeName, name, value) => _preventToString({
    astType: 'ConstVar',
    name,
    type: typeName,
    value: _prepareVarValue(value),
});
const Func = (name, args = [], returnType = 'void') => (strings, ...content) => _preventToString({
    astType: 'Func',
    name,
    args,
    returnType,
    body: ast(strings, ...content),
});
const AnonFunc = (args = [], returnType = 'void') => (strings, ...content) => _preventToString({
    astType: 'Func',
    name: null,
    args,
    returnType,
    body: ast(strings, ...content),
});
const Class = (name, members) => _preventToString({
    astType: 'Class',
    name,
    members,
});
const Sequence = (content) => ({
    astType: 'Sequence',
    content: _processRawContent(_intersperse(content, countTo(content.length - 1).map(() => '\n'))),
});
const ast = (strings, ...content) => _preventToString({
    astType: 'Sequence',
    content: _processRawContent(_intersperse(strings, content)),
});
const _processRawContent = (content) => {
    // 1. Flatten arrays and AstSequence, filter out nulls, and convert numbers to strings
    // Basically converts input to an Array<AstContent>.
    const flattenedAndFiltered = content.flatMap((element) => {
        if (typeof element === 'string') {
            return [element];
        }
        else if (typeof element === 'number') {
            return [element.toString()];
        }
        else {
            if (element === null) {
                return [];
            }
            else if (Array.isArray(element)) {
                return _processRawContent(_intersperse(element, countTo(element.length - 1).map(() => '\n')));
            }
            else if (typeof element === 'object' &&
                element.astType === 'Sequence') {
                return element.content;
            }
            else {
                return [element];
            }
        }
    });
    // 2. Combine adjacent strings
    const [combinedContent, remainingString] = flattenedAndFiltered.reduce(([combinedContent, currentString], element) => {
        if (typeof element === 'string') {
            return [combinedContent, currentString + element];
        }
        else {
            if (currentString.length) {
                return [[...combinedContent, currentString, element], ''];
            }
            else {
                return [[...combinedContent, element], ''];
            }
        }
    }, [[], '']);
    if (remainingString.length) {
        combinedContent.push(remainingString);
    }
    return combinedContent;
};
/**
 * Intersperse content from array1 with content from array2.
 * `array1.length` must be equal to `array2.length + 1`.
 */
const _intersperse = (array1, array2) => {
    if (array1.length === 0) {
        return [];
    }
    return array1
        .slice(1)
        .reduce((combinedContent, element, i) => {
        return combinedContent.concat([array2[i], element]);
    }, [array1[0]]);
};
/**
 * Prevents AST elements from being rendered as a string, as this is
 * most likely an error due to unproper use of `ast`.
 * Deacivated. Activate for debugging by uncommenting the line below.
 */
const _preventToString = (element) => ({
    ...element,
    // Uncomment this to activate
    // toString: () => { throw new Error(`Rendering element ${elemennt.astType} as string is probably an error`) }
});
const _prepareVarValue = (value) => {
    if (typeof value === 'number') {
        return Sequence([value.toString()]);
    }
    else if (typeof value === 'string') {
        return Sequence([value]);
    }
    else {
        return value;
    }
};

export { AnonFunc, Class, ConstVar, Func, Sequence, Var, _processRawContent, ast };
